@@ -9,10 +9,12 @@ use solana_sdk::transaction::Transaction;
 use {
     crate::{
         cuda_runtime::PinnedVec,
-        packet::{Packet, Packets},
+        packet::{Packet, PacketBatch},
         perf_libs,
         recycler::Recycler,
     },
+    ahash::AHasher,
+    rand::{thread_rng, Rng},
     rayon::ThreadPool,
     solana_metrics::inc_new_counter_debug,
     solana_rayon_threadlimit::get_thread_count,
@@ -20,6 +22,9 @@ use {
         hash::Hash, message::MESSAGE_HEADER_LENGTH, pubkey::Pubkey, short_vec::decode_shortu16_len,
         signature::Signature,
     },
+    std::hash::Hasher,
+    std::sync::atomic::{AtomicBool, AtomicU64, Ordering},
+    std::time::{Duration, Instant},
     std::{convert::TryFrom, mem::size_of},
 };
 
@@ -156,8 +161,8 @@ fn verify_packet(packet: &mut Packet, reject_non_vote: bool) {
     }
 }
 
-pub fn batch_size(batches: &[Packets]) -> usize {
-    batches.iter().map(|p| p.packets.len()).sum()
+pub fn count_packets_in_batches(batches: &[PacketBatch]) -> usize {
+    batches.iter().map(|batch| batch.packets.len()).sum()
 }
 
 // internal function to be unit-tested; should be used only by get_packet_offsets
@@ -336,7 +341,7 @@ fn check_for_simple_vote_transaction(
 }
 
 pub fn generate_offsets(
-    batches: &mut [Packets],
+    batches: &mut [PacketBatch],
     recycler: &Recycler<TxOffset>,
     reject_non_vote: bool,
 ) -> TxOffsets {
@@ -351,9 +356,9 @@ pub fn generate_offsets(
     msg_sizes.set_pinnable();
     let mut current_offset: usize = 0;
     let mut v_sig_lens = Vec::new();
-    batches.iter_mut().for_each(|p| {
+    batches.iter_mut().for_each(|batch| {
         let mut sig_lens = Vec::new();
-        p.packets.iter_mut().for_each(|packet| {
+        batch.packets.iter_mut().for_each(|packet| {
             let packet_offsets = get_packet_offsets(packet, current_offset, reject_non_vote);
 
             sig_lens.push(packet_offsets.sig_len);
@@ -388,30 +393,104 @@ pub fn generate_offsets(
     )
 }
 
-pub fn ed25519_verify_cpu(batches: &mut [Packets], reject_non_vote: bool) {
-    use rayon::prelude::*;
-    let count = batch_size(batches);
-    debug!("CPU ECDSA for {}", batch_size(batches));
-    PAR_THREAD_POOL.install(|| {
-        batches.into_par_iter().for_each(|p| {
-            p.packets
-                .par_iter_mut()
-                .for_each(|p| verify_packet(p, reject_non_vote))
-        })
-    });
-    inc_new_counter_debug!("ed25519_verify_cpu", count);
+pub struct Deduper {
+    filter: Vec<AtomicU64>,
+    seed: (u128, u128),
+    age: Instant,
+    max_age: Duration,
+    pub saturated: AtomicBool,
 }
 
-pub fn ed25519_verify_disabled(batches: &mut [Packets]) {
+impl Deduper {
+    pub fn new(size: u32, max_age_ms: u64) -> Self {
+        let mut filter: Vec<AtomicU64> = Vec::with_capacity(size as usize);
+        filter.resize_with(size as usize, Default::default);
+        let seed = thread_rng().gen();
+        Self {
+            filter,
+            seed,
+            age: Instant::now(),
+            max_age: Duration::from_millis(max_age_ms),
+            saturated: AtomicBool::new(false),
+        }
+    }
+
+    pub fn reset(&mut self) {
+        let now = Instant::now();
+        //this should reset every 500k unique packets per 1m sized deduper
+        //false positive rate is 1/1000 at that point
+        let saturated = self.saturated.load(Ordering::Relaxed);
+        if saturated || now.duration_since(self.age) > self.max_age {
+            for i in &self.filter {
+                i.store(0, Ordering::Relaxed);
+            }
+            self.seed = thread_rng().gen();
+            self.age = now;
+            self.saturated.store(false, Ordering::Relaxed);
+        }
+    }
+
+    fn dedup_packet(&self, count: &AtomicU64, packet: &mut Packet) {
+        // If this packet was already marked as discard, drop it
+        if packet.meta.discard {
+            return;
+        }
+        let mut hasher = AHasher::new_with_keys(self.seed.0, self.seed.1);
+        hasher.write(&packet.data[0..packet.meta.size]);
+        let hash = hasher.finish();
+        let len = self.filter.len();
+        let pos = (usize::try_from(hash).unwrap()).wrapping_rem(len);
+        // saturate each position with or
+        let prev = self.filter[pos].fetch_or(hash, Ordering::Relaxed);
+        if prev == u64::MAX {
+            self.saturated.store(true, Ordering::Relaxed);
+            //reset this value
+            self.filter[pos].store(hash, Ordering::Relaxed);
+        }
+        if hash == prev & hash {
+            packet.meta.discard = true;
+            count.fetch_add(1, Ordering::Relaxed);
+        }
+    }
+
+    pub fn dedup_packets(&self, batches: &mut [PacketBatch]) -> u64 {
+        let count = AtomicU64::new(0);
+        batches.iter_mut().for_each(|batch| {
+            batch
+                .packets
+                .iter_mut()
+                .for_each(|p| self.dedup_packet(&count, p))
+        });
+        count.load(Ordering::Relaxed)
+    }
+}
+
+pub fn ed25519_verify_cpu(batches: &mut [PacketBatch], reject_non_vote: bool) {
     use rayon::prelude::*;
-    let count = batch_size(batches);
-    debug!("disabled ECDSA for {}", batch_size(batches));
-    batches.into_par_iter().for_each(|p| {
-        p.packets
+    let packet_count = count_packets_in_batches(batches);
+    debug!("CPU ECDSA for {}", packet_count);
+    PAR_THREAD_POOL.install(|| {
+        batches.into_par_iter().for_each(|batch| {
+            batch
+                .packets
+                .par_iter_mut()
+                .for_each(|p| verify_packet(p, reject_non_vote))
+        });
+    });
+    inc_new_counter_debug!("ed25519_verify_cpu", packet_count);
+}
+
+pub fn ed25519_verify_disabled(batches: &mut [PacketBatch]) {
+    use rayon::prelude::*;
+    let packet_count = count_packets_in_batches(batches);
+    debug!("disabled ECDSA for {}", packet_count);
+    batches.into_par_iter().for_each(|batch| {
+        batch
+            .packets
             .par_iter_mut()
             .for_each(|p| p.meta.discard = false)
     });
-    inc_new_counter_debug!("ed25519_verify_disabled", count);
+    inc_new_counter_debug!("ed25519_verify_disabled", packet_count);
 }
 
 pub fn copy_return_values(sig_lens: &[Vec<u32>], out: &PinnedVec<u8>, rvs: &mut Vec<Vec<u8>>) {
@@ -465,7 +544,7 @@ pub fn get_checked_scalar(scalar: &[u8; 32]) -> Result<[u8; 32], PacketError> {
     Ok(out)
 }
 
-pub fn mark_disabled(batches: &mut [Packets], r: &[Vec<u8>]) {
+pub fn mark_disabled(batches: &mut [PacketBatch], r: &[Vec<u8>]) {
     batches.iter_mut().zip(r).for_each(|(b, v)| {
         b.packets.iter_mut().zip(v).for_each(|(p, f)| {
             p.meta.discard = *f == 0;
@@ -474,7 +553,7 @@ pub fn mark_disabled(batches: &mut [Packets], r: &[Vec<u8>]) {
 }
 
 pub fn ed25519_verify(
-    batches: &mut [Packets],
+    batches: &mut [PacketBatch],
     recycler: &Recycler<TxOffset>,
     recycler_out: &Recycler<PinnedVec<u8>>,
     reject_non_vote: bool,
@@ -486,21 +565,21 @@ pub fn ed25519_verify(
     let api = api.unwrap();
 
     use crate::packet::PACKET_DATA_SIZE;
-    let count = batch_size(batches);
+    let packet_count = count_packets_in_batches(batches);
 
     // micro-benchmarks show GPU time for smallest batch around 15-20ms
     // and CPU speed for 64-128 sigverifies around 10-20ms. 64 is a nice
     // power-of-two number around that accounting for the fact that the CPU
     // may be busy doing other things while being a real validator
     // TODO: dynamically adjust this crossover
-    if count < 64 {
+    if packet_count < 64 {
         return ed25519_verify_cpu(batches, reject_non_vote);
     }
 
     let (signature_offsets, pubkey_offsets, msg_start_offsets, msg_sizes, sig_lens) =
         generate_offsets(batches, recycler, reject_non_vote);
 
-    debug!("CUDA ECDSA for {}", batch_size(batches));
+    debug!("CUDA ECDSA for {}", packet_count);
     debug!("allocating out..");
     let mut out = recycler_out.allocate("out_buffer");
     out.set_pinnable();
@@ -508,15 +587,15 @@ pub fn ed25519_verify(
     let mut rvs = Vec::new();
 
     let mut num_packets: usize = 0;
-    for p in batches.iter() {
+    for batch in batches.iter() {
         elems.push(perf_libs::Elems {
-            elems: p.packets.as_ptr(),
-            num: p.packets.len() as u32,
+            elems: batch.packets.as_ptr(),
+            num: batch.packets.len() as u32,
         });
         let mut v = Vec::new();
-        v.resize(p.packets.len(), 0);
+        v.resize(batch.packets.len(), 0);
         rvs.push(v);
-        num_packets = num_packets.saturating_add(p.packets.len());
+        num_packets = num_packets.saturating_add(batch.packets.len());
     }
     out.resize(signature_offsets.len(), 0);
     trace!("Starting verify num packets: {}", num_packets);
@@ -545,7 +624,7 @@ pub fn ed25519_verify(
     trace!("done verify");
     copy_return_values(&sig_lens, &out, &mut rvs);
     mark_disabled(batches, &rvs);
-    inc_new_counter_debug!("ed25519_verify_gpu", count);
+    inc_new_counter_debug!("ed25519_verify_gpu", packet_count);
 }
 
 #[cfg(test)]
@@ -565,7 +644,7 @@ mod tests {
     use {
         super::*,
         crate::{
-            packet::{Packet, Packets},
+            packet::{to_packet_batches, Packet, PacketBatch},
             sigverify::{self, PacketOffsets},
             test_tx::{test_multisig_tx, test_tx, vote_tx},
         },
@@ -593,9 +672,9 @@ mod tests {
 
     #[test]
     fn test_mark_disabled() {
-        let mut batch = Packets::default();
+        let mut batch = PacketBatch::default();
         batch.packets.push(Packet::default());
-        let mut batches: Vec<Packets> = vec![batch];
+        let mut batches: Vec<PacketBatch> = vec![batch];
         mark_disabled(&mut batches, &[vec![0]]);
         assert!(batches[0].packets[0].meta.discard);
         mark_disabled(&mut batches, &[vec![1]]);
@@ -717,7 +796,7 @@ mod tests {
         assert!(packet.meta.discard);
 
         packet.meta.discard = false;
-        let mut batches = generate_packet_vec(&packet, 1, 1);
+        let mut batches = generate_packet_batches(&packet, 1, 1);
         ed25519_verify(&mut batches);
         assert!(batches[0].packets[0].meta.discard);
     }
@@ -753,7 +832,7 @@ mod tests {
         assert!(packet.meta.discard);
 
         packet.meta.discard = false;
-        let mut batches = generate_packet_vec(&packet, 1, 1);
+        let mut batches = generate_packet_batches(&packet, 1, 1);
         ed25519_verify(&mut batches);
         assert!(batches[0].packets[0].meta.discard);
     }
@@ -878,21 +957,21 @@ mod tests {
         );
     }
 
-    fn generate_packet_vec(
+    fn generate_packet_batches(
         packet: &Packet,
         num_packets_per_batch: usize,
         num_batches: usize,
-    ) -> Vec<Packets> {
+    ) -> Vec<PacketBatch> {
         // generate packet vector
         let batches: Vec<_> = (0..num_batches)
             .map(|_| {
-                let mut packets = Packets::default();
-                packets.packets.resize(0, Packet::default());
+                let mut packet_batch = PacketBatch::default();
+                packet_batch.packets.resize(0, Packet::default());
                 for _ in 0..num_packets_per_batch {
-                    packets.packets.push(packet.clone());
+                    packet_batch.packets.push(packet.clone());
                 }
-                assert_eq!(packets.packets.len(), num_packets_per_batch);
-                packets
+                assert_eq!(packet_batch.packets.len(), num_packets_per_batch);
+                packet_batch
             })
             .collect();
         assert_eq!(batches.len(), num_batches);
@@ -909,7 +988,7 @@ mod tests {
             packet.data[20] = packet.data[20].wrapping_add(10);
         }
 
-        let mut batches = generate_packet_vec(&packet, n, 2);
+        let mut batches = generate_packet_batches(&packet, n, 2);
 
         // verify packets
         ed25519_verify(&mut batches);
@@ -918,11 +997,11 @@ mod tests {
         let should_discard = modify_data;
         assert!(batches
             .iter()
-            .flat_map(|p| &p.packets)
+            .flat_map(|batch| &batch.packets)
             .all(|p| p.meta.discard == should_discard));
     }
 
-    fn ed25519_verify(batches: &mut [Packets]) {
+    fn ed25519_verify(batches: &mut [PacketBatch]) {
         let recycler = Recycler::default();
         let recycler_out = Recycler::default();
         sigverify::ed25519_verify(batches, &recycler, &recycler_out, false);
@@ -935,13 +1014,13 @@ mod tests {
         tx.signatures.pop();
         let packet = sigverify::make_packet_from_transaction(tx);
 
-        let mut batches = generate_packet_vec(&packet, 1, 1);
+        let mut batches = generate_packet_batches(&packet, 1, 1);
 
         // verify packets
         ed25519_verify(&mut batches);
         assert!(batches
             .iter()
-            .flat_map(|p| &p.packets)
+            .flat_map(|batch| &batch.packets)
             .all(|p| p.meta.discard));
     }
 
@@ -969,7 +1048,7 @@ mod tests {
 
         let n = 4;
         let num_batches = 3;
-        let mut batches = generate_packet_vec(&packet, n, num_batches);
+        let mut batches = generate_packet_batches(&packet, n, num_batches);
 
         packet.data[40] = packet.data[40].wrapping_add(8);
 
@@ -984,7 +1063,7 @@ mod tests {
         ref_vec[0].push(0u8);
         assert!(batches
             .iter()
-            .flat_map(|p| &p.packets)
+            .flat_map(|batch| &batch.packets)
             .zip(ref_vec.into_iter().flatten())
             .all(|(p, discard)| {
                 if discard == 0 {
@@ -1008,7 +1087,7 @@ mod tests {
         for _ in 0..50 {
             let n = thread_rng().gen_range(1, 30);
             let num_batches = thread_rng().gen_range(2, 30);
-            let mut batches = generate_packet_vec(&packet, n, num_batches);
+            let mut batches = generate_packet_batches(&packet, n, num_batches);
 
             let num_modifications = thread_rng().gen_range(0, 5);
             for _ in 0..num_modifications {
@@ -1029,8 +1108,8 @@ mod tests {
             // check result
             batches
                 .iter()
-                .flat_map(|p| &p.packets)
-                .zip(batches_cpu.iter().flat_map(|p| &p.packets))
+                .flat_map(|batch| &batch.packets)
+                .zip(batches_cpu.iter().flat_map(|batch| &batch.packets))
                 .for_each(|(p1, p2)| assert_eq!(p1, p2));
         }
     }
@@ -1182,7 +1261,7 @@ mod tests {
         solana_logger::setup();
 
         let mut current_offset = 0usize;
-        let mut batch = Packets::default();
+        let mut batch = PacketBatch::default();
         batch
             .packets
             .push(sigverify::make_packet_from_transaction(test_tx()));
@@ -1206,4 +1285,62 @@ mod tests {
                 current_offset = current_offset.saturating_add(size_of::<Packet>());
             });
     }
+
+    #[test]
+    fn test_dedup_same() {
+        let tx = test_tx();
+
+        let mut batches =
+            to_packet_batches(&std::iter::repeat(tx).take(1024).collect::<Vec<_>>(), 128);
+        let packet_count = sigverify::count_packets_in_batches(&batches);
+        let filter = Deduper::new(1_000_000, 0);
+        let discard = filter.dedup_packets(&mut batches) as usize;
+        assert_eq!(packet_count, discard + 1);
+    }
+
+    #[test]
+    fn test_dedup_diff() {
+        let mut filter = Deduper::new(1_000_000, 0);
+        let mut batches = to_packet_batches(&(0..1024).map(|_| test_tx()).collect::<Vec<_>>(), 128);
+
+        let discard = filter.dedup_packets(&mut batches) as usize;
+        // because dedup uses a threadpool, there maybe up to N threads of txs that go through
+        assert_eq!(discard, 0);
+        filter.reset();
+        for i in filter.filter {
+            assert_eq!(i.load(Ordering::Relaxed), 0);
+        }
+    }
+
+    #[test]
+    #[ignore]
+    fn test_dedup_saturated() {
+        let filter = Deduper::new(1_000_000, 0);
+        let mut discard = 0;
+        assert!(!filter.saturated.load(Ordering::Relaxed));
+        for i in 0..1000 {
+            let mut batches =
+                to_packet_batches(&(0..1000).map(|_| test_tx()).collect::<Vec<_>>(), 128);
+            discard += filter.dedup_packets(&mut batches) as usize;
+            println!("{} {}", i, discard);
+            if filter.saturated.load(Ordering::Relaxed) {
+                break;
+            }
+        }
+        assert!(filter.saturated.load(Ordering::Relaxed));
+    }
+
+    #[test]
+    fn test_dedup_false_positive() {
+        let filter = Deduper::new(1_000_000, 0);
+        let mut discard = 0;
+        for i in 0..10 {
+            let mut batches =
+                to_packet_batches(&(0..1024).map(|_| test_tx()).collect::<Vec<_>>(), 128);
+            discard += filter.dedup_packets(&mut batches) as usize;
+            println!("false positive rate: {}/{}", discard, i * 1024);
+        }
+        //allow for 1 false positive even if extremely unlikely
+        assert!(discard < 2);
+    }
 }
